# HealMates release 4 by David Bowland
# ./addons/eventscripts/healmates/healmates.py

# >>> To configure this addon please see healmates.cfg <<<

"""
Allows players to heal surrounding teammates. Compatible with bots! Requires only ES 2.0+
"""


import es
import gamethread
import langlib
import os.path
import playerlib
import random


info = es.AddonInfo()
info.name = 'HealMates'
info.version = '4'
info.url = 'http://addons.eventscripts.com/addons/view/healmates'
info.basename = 'healmates'
info.author = 'SuperDave'


str_last_sound = ''
dict_say_commands = {'purchase':'', 'heal':''}
dict_healing_players = {}

tuple_sounds = ('vo/npc/male01/health01.wav', 'vo/npc/male01/health02.wav', 'vo/npc/male01/health03.wav', 'vo/npc/male01/health04.wav', 'vo/npc/male01/health05.wav')
dict_options = {'healmates_ability_cost':[2000, 'Cost of the ability to heal teammates--set to 0 for players to automatically receive the healing ability'], 'healmates_ability_buyzone':[1, '0 = players can buy the ability to heal teammates anywhere, 1 = players must be in a buy zone to purchase the ability to heal teammates'], 'healmates_ability_trigger':['!buyheal' ,'Command players must say to purchase the ability to heal teammates'], 'healmates_ability_loseondeath':[1, '0 = players keep the ability to heal players when killed, 1 = players lose the ability to heal players when killed'], 'healmates_heal_cost':[10, 'Cost of healing another player one health point--set to 0 for no cost'], 'healmates_heal_maxgain':[30, 'Maximum amount of health a player can gain from being healed'], 'healmates_heal_maxendhealth':[85, 'Maximum amount of health a player can be healed to'], 'healmates_heal_distance':[200, 'Maximum distance a player can be from a healing player to be healed'], 'healmates_heal_distance_percent':[65, 'Percent of the health possible to gain from healing a player at the maximum distance will gain--set to 100 for players at the maximum healing distance to be healed the maximum amount'], 'healmates_heal_autoheal':[0, '0 = players must trigger healing of teammates, 1 = healing of teammates is handled automatically'], 'healmates_heal_sound':[35, 'Percent of the map that can hear when a player heals teammates--set to 0 for no sound'], 'healmates_heal_trigger':['!heal', 'Command players must say to heal teammates'], 'healmates_heal_trigger_attemptdelay':[4, 'Number of seconds required between heal attempts'], 'healmates_heal_autoheal_delay':[6, 'Number of seconds between forcing players with the healing ability to heal surrounding teammates--set to 0 to eliminate'], 'healmates_heal_autoheal_ecolimit':[6000, 'Minimum amount of money player will have after automatically healing'], 'healmates_bot':[1, '0 = bots will neither buy the healing ability nor heal teammates randomly, 1 = bots will buy the healing ability and heal teammates randomly'], 'healmates_bot_buychance':[70, 'Percent chance that a bot with enough money will buy the ability to heal teammates'], 'healmates_bot_ecolimit_buy':[8000, 'Amount of money a bot must have to attempt buying the healing ability'], 'healmates_bot_ecolimit_heal':[4000, 'Amount of money bots will refrain from spending on healing teammates'], 'healmates_bot_delay_min':[4, 'Minimum number of seconds from healing teammates until a bot randomly heals teammates again'], 'healmates_bot_delay_max':[25, 'Maximum number of seconds from healing teammates until a bot randomly heals teammates again'], 'healmates_announce_ability':[1, '0 = no change, 1 = announce command to purchase the healing ability each spawn'], 'healmates_announce_ability_feedback':[2, '0 = no change, 1 = announce feedback from purchasing the healing ability in chat area, 2 = announce feedback from purchasing the healing ability in center text'], 'healmates_announce_heal':[2, '0 = no change, 1 = announce command to heal teammates only when a player acquires the healing ability, 2 = announce command to heal teammates each spawn'], 'healmates_announce_heal_feedback':[2, '0 = no change, 1 = announce error feedback from healing teammates in chat area, 2 = announce error feedback from healing teammates in center text'], 'healmates_announce_heal_healer':[2, '0 = no change, 1 = announce to players when they heal another player in chat area, 2 = announce to players when they heal another player on the console'], 'healmates_announce_heal_healed':[1, '0 = no change, 1 = announce to players when they were healed by another player in chat area, 2 = announce to players when they were healed by another player on the console, 3 = announce to players when they were healed by another player in center text'], 'healmates_announce_heal_console':[1, '0 = no change, 1 = announce to players on the console when a teammates is healed, 2 = announce to players on the console when any player is healed']}
func_lang_text = None


def load():
   """
   Checks to make sure critical server variables are created by the config
   Creates say commands for triggering healing
   """
   global dict_say_commands
   global func_lang_text

   for str_option in dict_options:
      es.ServerVar(str_option, dict_options[str_option][0], dict_options[str_option][1])
   if os.path.isfile(es.getAddonPath('healmates') + '/healmates.cfg'):
      es.server.cmd('es_xmexec ../addons/eventscripts/healmates/healmates.cfg')
   else:
      es.dbgmsg(0, 'HealMates: Unable to load healmates.cfg! Please ensure it is in the ./healmates/ directory.')

   str_lang_path = es.getAddonPath('healmates') + '/healmates_languages.ini'
   if os.path.isfile(str_lang_path):
      func_lang_text = langlib.Strings(str_lang_path)
   else:
      es.dbgmsg(0, 'HealMates: Unable to load healmates_languages.ini! Please ensure it is in the ./healmates/ directory.')
   if not func_lang_text:
      func_lang_text = missing_lang

   str_command = str(es.ServerVar('healmates_ability_trigger'))
   if str_command:
      if not es.exists('saycommand', str_command):
         es.regsaycmd(str_command, 'healmates/buy_cmd', 'Say command for buying the ability to heal teammates')
         dict_say_commands['purchase'] = str_command

   str_command = str(es.ServerVar('healmates_heal_trigger'))
   if str_command:
      if not es.exists('saycommand', str_command):
         es.regsaycmd(str_command, 'healmates/heal_cmd', 'Say command for healing surrounding teammates')
         dict_say_commands['heal'] = str_command


def es_map_start(event_var):
   """Removes outstanding delays"""
   global dict_healing_players

   for int_userid in dict_healing_players:
      dict_current_player = dict_healing_players[int_userid]
      if dict_current_player['heal_limit'] or dict_current_player['auto_heal']:
         gamethread.cancelDelayed('healmates_%s' % int_userid)
      if dict_current_player['bot_delay']:
         gamethread.cancelDelayed('healmates_bot_%s' % int_userid)

   dict_healing_players.clear()


def round_start(event_var):
   """Removes outstanding delays"""
   global dict_healing_players

   int_auto_heal = int(es.ServerVar('healmates_heal_autoheal'))
   for int_userid in dict_healing_players:
      dict_current_player = dict_healing_players[int_userid]
      if dict_current_player['heal_limit'] > 0:
         gamethread.cancelDelayed('healmates_%s' % int_userid)
         dict_current_player['heal_limit'] = 0
      if int_auto_heal and not dict_current_player['auto_heal']:
         gamethread.delayedname(es.ServerVar('healmates_heal_autoheal_delay'), 'healmates_autoheal_%s' % int_userid, auto_heal, int_userid)
         dict_current_player['auto_heal'] = 1
         dict_current_player['heal_limit'] = - 1


def player_spawn(event_var):
   """
   Displays the message telling players how to buy or use the healing ability
   Handles bots buying the healing ability
   """
   if int(event_var['es_userteam']) < 2:
      return

   int_userid = int(event_var['userid'])
   int_cost = int(es.ServerVar('healmates_ability_cost'))
   str_player_lang = playerlib.getPlayer(int_userid).get('lang')
   if not int_cost and not dict_healing_players.has_key(int_userid):
      give_ability(int_userid, str_player_lang)

   if dict_healing_players.has_key(int_userid):
      int_announce = int(es.ServerVar('healmates_announce_heal'))
      if int_announce and int_announce <> 1:
         es.tell(int_userid, '#multi', func_lang_text('heal teammates', dict_say_commands, str_player_lang))
         int_cost = int(es.ServerVar('healmates_heal_cost'))
         if int_cost:
            es.tell(int_userid, '#multi', func_lang_text('heal cost', {'cost':int_cost}, str_player_lang))

   else:
      if event_var['es_steamid'] == 'BOT' and int(es.ServerVar('healmates_bot')):
         int_money = es.getplayerprop(int_userid, 'CCSPlayer.m_iAccount')
         if int_money >= es.ServerVar('healmates_bot_ecolimit_buy') and int_money >= int_cost and random.randint(1, 100) <= es.ServerVar('healmates_bot_buychance') and (es.getplayerprop(int_userid, 'CCSPlayer.m_bInBuyZone') or not int(es.ServerVar('healmates_ability_buyzone'))):
            es.setplayerprop(int_userid, 'CCSPlayer.m_iAccount', int_money - int_cost)
            give_ability(int_userid, str_player_lang)
      elif int(es.ServerVar('healmates_announce_ability')):
         es.tell(int_userid, '#multi', func_lang_text('ability purchase', {'purchase':dict_say_commands['purchase'], 'cost':int_cost}, str_player_lang))


def player_death(event_var):
   """Removes delays for the dead player"""
   global dict_healing_players

   int_userid = int(event_var['userid'])
   if dict_healing_players.has_key(int_userid):
      dict_current_player = dict_healing_players[int_userid]
      if dict_current_player['heal_limit'] or dict_current_player['auto_heal']:
         gamethread.cancelDelayed('healmates_%s' % int_userid)
      if dict_current_player['bot_delay']:
         gamethread.cancelDelayed('healmates_bot_%s' % int_userid)
      if int(es.ServerVar('healmates_ability_loseondeath')):
         del dict_healing_players[int_userid]


def player_disconnect(event_var):
   """Removes delays and dictionary element for the dead player"""
   global dict_healing_players

   int_userid = int(event_var['userid'])
   if dict_healing_players.has_key(int_userid):
      dict_current_player = dict_healing_players[int_userid]
      if dict_current_player['heal_limit'] or dict_current_player['auto_heal']:
         gamethread.cancelDelayed('healmates_%s' % int_userid)
      if dict_current_player['bot_delay']:
         gamethread.cancelDelayed('healmates_bot_%s' % int_userid)
      del dict_healing_players[int_userid]


def unload():
   """
   Calls es_map_start
   Unregisters say commands
   """
   es_map_start({})

   for str_command in dict_say_commands:
      if dict_say_commands[str_command]:
         es.unregsaycmd(dict_say_commands[str_command])

   for str_option in dict_options:
      es.ServerVar(str_option).set(0)


def buy_cmd():
   """Calls give_ability for the player buying the healing ability"""
   int_userid = es.getcmduserid()
   str_player_lang = playerlib.getPlayer(int_userid).get('lang')
   int_feedback = int(es.ServerVar('healmates_announce_ability_feedback'))
   if es.getplayerteam(int_userid) > 1 and not es.getplayerprop(int_userid, 'CBasePlayer.pl.deadflag'):
      if not dict_healing_players.has_key(int_userid):
         if es.getplayerprop(int_userid, 'CCSPlayer.m_bInBuyZone') or not int(es.ServerVar('healmates_ability_buyzone')):
            int_cost = int(es.ServerVar('healmates_ability_cost'))
            int_money = es.getplayerprop(int_userid, 'CCSPlayer.m_iAccount')
            if int_money >= int_cost:
               es.setplayerprop(int_userid, 'CCSPlayer.m_iAccount', int_money - int_cost)

               give_ability(int_userid, str_player_lang)

            else:
               command_feedback(int_userid, 'ability cash', {'cash':int_cost}, str_player_lang, int_feedback)
         else:
            command_feedback(int_userid, 'ability buy zone', {}, str_player_lang, int_feedback)
      else:
         command_feedback(int_userid, 'ability twice', {}, str_player_lang, int_feedback)
   else:
      command_feedback(int_userid, 'ability dead', {}, str_player_lang, int_feedback)


def heal_cmd():
   """
   Triggers execute_heal for the player with the desired arguments
   Starts a delay on the player that calls remove_restriction so the player cannot use this command until the delay expires
   """
   global dict_healing_players

   int_userid = es.getcmduserid()
   str_player_lang = playerlib.getPlayer(int_userid).get('lang')
   int_feedback = int(es.ServerVar('healmates_announce_heal_feedback'))

   if dict_healing_players.has_key(int_userid):
      if es.getplayerteam(int_userid) > 1 and not es.getplayerprop(int_userid, 'CBasePlayer.pl.deadflag'):
         int_money = es.getplayerprop(int_userid, 'CCSPlayer.m_iAccount')
         if es.getargc() > 1:
            int_amount = int(es.getargv(1))
            if int_amount > int_money:
               int_amount = int_money
         else:
            int_amount = int_money

         int_unit_cost = int(es.ServerVar('healmates_heal_cost'))
         if int_amount >= int_unit_cost:
            if not dict_healing_players[int_userid]['heal_limit']:
               float_attempt_delay = float(es.ServerVar('healmates_heal_trigger_attemptdelay'))
               if execute_heal(int_userid, int_amount, int_unit_cost) and float_attempt_delay:
                  dict_healing_players[int_userid]['heal_limit'] = 1
                  if dict_healing_players[int_userid]['auto_heal']:
                     gamethread.cancelDelayed('healmates_%s' % int_userid)
                     gamethread.delayedname(float_attempt_delay, 'healmates_%s' % int_userid, auto_heal, int_userid)
                  else:
                     gamethread.delayedname(float_attempt_delay, 'healmates_%s' % int_userid, remove_restriction, int_userid)
               elif dict_healing_players[int_userid]['auto_heal']:
                  gamethread.cancelDelayed('healmates_%s' % int_userid)
                  gamethread.delayedname(es.ServerVar('healmates_heal_autoheal_delay'), 'healmates_%s' % int_userid, remove_restriction, int_userid)

            else:
               command_feedback(int_userid, 'heal limit', {}, str_player_lang, int_feedback)
         else:
            command_feedback(int_userid, 'heal cash', {'cash':int_unit_cost}, str_player_lang, int_feedback)
      else:
         command_feedback(int_userid, 'heal dead', {}, str_player_lang, int_feedback)
   else:
      command_feedback(int_userid, 'heal fail', {}, str_player_lang, int_feedback)


def give_ability(int_userid, str_player_lang):
   """
   Gives the player the ability to heal teammates
   Announces instructions on using the healing ability
   """
   global dict_healing_players

   int_auto_heal = int(es.ServerVar('healmates_heal_autoheal'))
   dict_healing_players[int_userid] = {'auto_heal':int_auto_heal, 'heal_limit':int_auto_heal * - 1, 'bot_delay':0}
   if int_auto_heal:
      gamethread.delayedname(es.ServerVar('healmates_heal_autoheal_delay'), 'healmates_%s' % int_userid, auto_heal, int_userid)
   elif es.getplayersteamid(int_userid) == 'BOT' and int(es.ServerVar('healmates_bot')):
      dict_healing_players[int_userid]['bot_delay'] = 1
      gamethread.delayedname(random.randint(int(es.ServerVar('healmates_bot_delay_min')), int(es.ServerVar('healmates_bot_delay_max'))), 'healmates_bot_%s' % int_userid, bot_heal, int_userid)

   command_feedback(int_userid, 'ability success', {}, str_player_lang, int(es.ServerVar('healmates_announce_ability_feedback')))
   int_announce = int(es.ServerVar('healmates_announce_heal'))
   if int_announce:
      es.tell(int_userid, '#multi', func_lang_text('heal teammates', dict_say_commands, str_player_lang))
      int_cost = int(es.ServerVar('healmates_heal_cost'))
      if int_cost:
         es.tell(int_userid, '#multi', func_lang_text('heal cost', {'cost':int_cost}, str_player_lang))


def execute_heal(int_userid, int_amount, int_unit_cost):
   """
   Gives health to teammates around the player
   Takes the money for the health given
   Emits a sound from the player
   Returns: 0 = no health given to teammates, 1 = health given to teammates
   """
   global str_last_sound

   bool_gave_health = 0
   int_team = es.getplayerteam(int_userid)
   dict_near_players = {}
   dict_player_list = es.createplayerlist()
   dict_current_player = dict_player_list[int_userid]
   for int_loop_userid in dict_player_list:
      dict_loop_player = dict_player_list[int_loop_userid]
      if int_team == int(dict_loop_player['teamid']) and int_userid <> int_loop_userid and not int(dict_loop_player['isdead']):
         float_distance = pow(pow(float(dict_current_player['x']) - float(dict_loop_player['x']), 2) + pow(float(dict_current_player['y']) - float(dict_loop_player['y']), 2) + pow(float(dict_current_player['z']) - float(dict_loop_player['z']), 2), 0.5)
         float_heal_distance = float(es.ServerVar('healmates_heal_distance'))
         if float_distance <= float_heal_distance:
            int_start_percent = int(es.ServerVar('healmates_heal_distance_percent'))
            int_max_gain = int(float(float(100 - int_start_percent) * (float_heal_distance - float_distance) / float_heal_distance + int_start_percent) * es.ServerVar('healmates_heal_maxgain') / 100)
            if int_max_gain:
               dict_near_players[int_loop_userid] = {'possible':int_max_gain, 'given':0}

   int_amount_spent = 0
   int_lowest_userid = -1
   int_highest_regen_health = int(es.ServerVar('healmates_heal_maxendhealth'))
   while (int_amount_spent + int_unit_cost <= int_amount or not int_unit_cost) and int_lowest_userid:
      int_lowest_health = int_highest_regen_health
      int_lowest_userid = 0
      for int_loop_userid in dict_near_players:
         dict_loop_player = dict_near_players[int_loop_userid]
         int_health = es.getplayerprop(int_loop_userid, 'CBasePlayer.m_iHealth')
         if int_health < int_lowest_health and dict_loop_player['possible'] > 0:
            int_lowest_userid = int_loop_userid
            int_lowest_health = int_health

      if int_lowest_userid:
         dict_current_player = dict_near_players[int_lowest_userid]
         dict_current_player['possible'] -= 1
         dict_current_player['given'] += 1
         es.setplayerprop(int_lowest_userid, 'CBasePlayer.m_iHealth', int_lowest_health + 1)
         int_amount_spent += int_unit_cost
         bool_gave_health = 1

   if int_amount_spent or not int_unit_cost:
      if int_amount_spent:
         es.setplayerprop(int_userid, 'CCSPlayer.m_iAccount', es.getplayerprop(int_userid, 'CCSPlayer.m_iAccount') - int_amount_spent)
      int_announce_healer = int(es.ServerVar('healmates_announce_heal_healer'))
      int_announce_healed = int(es.ServerVar('healmates_announce_heal_healed'))
      int_announce_console = int(es.ServerVar('healmates_announce_heal_console'))
      if int_announce_healer or int_announce_healed or int_announce_console:
         str_healer_lang = playerlib.getPlayer(int_userid).get('lang')
         dict_tokens = {'player':'', 'health':0, 'healer':es.getplayername(int_userid), 'cost':0}
         if int_unit_cost:
            str_healer_text = 'healer cost'
         else:
            str_healer_text = 'healer no cost'

         for int_loop_userid in dict_near_players:
            int_health = dict_tokens['health'] = dict_near_players[int_loop_userid]['given']
            if int_health:
               dict_tokens['player'] = es.getplayername(int_loop_userid)
               dict_tokens['cost'] = int_health * int_unit_cost

               if int_announce_healer:
                  if int_announce_healer == 1:
                     es.tell(int_userid, '#multi', func_lang_text(str_healer_text, dict_tokens, str_healer_lang))
                  else:
                     es.cexec(int_userid, 'echo ' + remove_tags(func_lang_text(str_healer_text, dict_tokens, str_healer_lang)))

               if int_announce_healed:
                  str_loop_lang = playerlib.getPlayer(int_loop_userid).get('lang')
                  if int_announce_healed == 3:
                     es.centertell(int_loop_userid, remove_tags(func_lang_text('healed', dict_tokens, str_loop_lang)))
                  elif int_announce_healed == 2:
                     es.cexec(int_loop_userid, 'echo ' + remove_tags(func_lang_text('healed', dict_tokens, str_loop_lang)))
                  else:
                     es.tell(int_loop_userid, '#multi', func_lang_text('healed', dict_tokens, str_loop_lang))

               if int_announce_console:
                  if int_announce_console == 2:
                     for int_console_userid in es.getUseridList():
                        es.cexec(int_console_userid, 'echo ' + func_lang_text('heal console', dict_tokens, playerlib.getPlayer(int_console_userid).get('lang')))
                  else:
                     for int_console_userid in dict_near_players:
                        if int_console_userid <> int_loop_userid:
                           es.cexec(int_console_userid, 'echo ' + func_lang_text('heal console', dict_tokens, playerlib.getPlayer(int_console_userid).get('lang')))

   float_sound = float(es.ServerVar('healmates_heal_sound'))
   if float_sound and bool_gave_health:
      list_possible_sounds = list(tuple_sounds)
      if str_last_sound in list_possible_sounds:
         list_possible_sounds.remove(str_last_sound)
      str_last_sound = random.choice(list_possible_sounds)
      es.emitsound('player', int_userid, str_last_sound, 1.0, (100 - float_sound) / 100)

   return bool_gave_health


def remove_restriction(int_userid):
   """Removes the heal delay restriction on the player"""
   global dict_healing_players

   if dict_healing_players.has_key(int_userid):
      dict_healing_players[int_userid]['heal_limit'] = 0


def auto_heal(int_userid):
   """Automatically executes the healing command on the player"""
   global dict_healing_players

   int_unit_cost = int(es.ServerVar('healmates_heal_cost'))
   int_money = es.getplayerprop(int_userid, 'CCSPlayer.m_iAccount') - int(es.ServerVar('healmates_heal_autoheal_ecolimit'))
   dict_current_player = dict_healing_players[int_userid]
   if int_money >= int_unit_cost or not int_unit_cost:
      if dict_current_player['heal_limit'] > 0:
         gamethread.cancelDelayed('healmates_%s' % int_userid)
         dict_current_player['heal_limit'] = -1
      if dict_current_player['bot_delay']:
         gamethread.cancelDelayed('healmates_bot_%s' % int_userid)
         dict_current_player['bot_delay'] = 0
      execute_heal(int_userid, int_money, int_unit_cost)

   else:
      dict_current_player['heal_limit'] = 0
      if es.getplayersteamid(int_userid) == 'BOT' and int(es.ServerVar('healmates_bot')):
         dict_current_player['bot_delay'] = 1
         gamethread.delayedname(random.randint(int(es.ServerVar('healmates_bot_delay_min')), int(es.ServerVar('healmates_bot_delay_max'))), 'healmates_bot_%s' % int_userid, bot_heal, int_userid)

   gamethread.delayedname(es.ServerVar('healmates_heal_autoheal_delay'), 'healmates_%s' % int_userid, auto_heal, int_userid)


def bot_heal(int_userid):
   """Randomly triggers the healing function for bots"""
   global dict_healing_players

   int_unit_cost = int(es.ServerVar('healmates_heal_cost'))
   int_money = es.getplayerprop(int_userid, 'CCSPlayer.m_iAccount') - int(es.ServerVar('healmates_bot_ecolimit_heal'))
   if int_money >= int_unit_cost or not int_unit_cost:
      dict_current_player = dict_healing_players[int_userid]
      if dict_current_player['auto_heal']:
         gamethread.cancelDelayed('healmates_%s' % int_userid)

      execute_heal(int_userid, int_money, int_unit_cost)

      if dict_current_player['auto_heal']:
         gamethread.delayedname(es.ServerVar('healmates_heal_autoheal_delay'), 'healmates_%s' % int_userid, auto_heal, int_userid)

   gamethread.delayedname(random.randint(int(es.ServerVar('healmates_bot_delay_min')), int(es.ServerVar('healmates_bot_delay_max'))), 'healmates_bot_%s' % int_userid, bot_heal, int_userid)


def command_feedback(int_userid, str_text, dict_tokens, str_lang, int_feedback):
   """Displays the message to the player in the specified langauge"""
   if int_feedback:
      if int_feedback == 2:
         es.centertell(int_userid, remove_tags(func_lang_text(str_text, dict_tokens, str_lang)))
      else:
         es.tell(int_userid, '#multi', func_lang_text(str_text, dict_tokens, str_lang))


def remove_tags(str_text):
   """Removes #lightgreen, #green and #default tags from the supplied string"""
   return str_text.replace('#lightgreen', '').replace('#green', '').replace('#default', '')


def missing_lang(str_heading, dict_tokens, str_language):
   """Displays an error message because the languages .ini could not be found"""
   return 'HealMates: healmates_languages.ini not found'